// Copyright 2006 all rights reserved; see LICENSE file for BSD-style license

package edu.berkeley.sbp;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;

/** <font color=green>juxtaposition; zero or more adjacent Elements; can specify a rewriting</font> */
public abstract class Sequence implements Iterable<Element>, SequenceOrElement {

    protected final Element[] elements;

    boolean needed_or_hated = false;
    boolean in_a_union = false;

    final HashSet<Sequence> needs  = new HashSet<Sequence>();
    final HashSet<Sequence> hates  = new HashSet<Sequence>();

    // FIXME: these are ugly -- migrate into Grammar
    HashMap<Sequence,Boolean> canNeed = new HashMap<Sequence,Boolean>();
    HashMap<Sequence,Boolean> canKill = new HashMap<Sequence,Boolean>();

    final Position          firstp;

    Atom follow = null;

    // Static Constructors //////////////////////////////////////////////////////////////////////////////

    /** create a sequence of one element */
    public static Sequence create(Element e) { return create(new Element[] { e }, 0); }

    /** create a sequence which drops the result of all but one of its element */
    public static Sequence create(Element[] e, int which) { return new Singleton(e, which); }

    /** create a sequence which always evaluates to a constant result  */
    public static Sequence create(Element[] e, Object result) { return new Constant(e, result); }

    /**
     *  create a sequence (general form)
     *  @param head   the head of the output tree
     *  @param e      the elements to match
     *  @param drop   only elements of <tt>e</tt> whose corresponding <tt>boolean</tt> in <tt>drops</tt>
     *                is <i>false</i> will be included in the output tree
     *  @param foster if true, all children of the last child (ie
     *                grandchildren) are promoted to children of this
     *                node; this is very useful for matching repetitions
     **/
    public static Sequence create(Object head, Element[] e, boolean[] drop, boolean foster) {
        return foster
            ? new Unwrap(e, head, drop)
            : new RewritingSequence(head, e, drop);
    }
    public static Sequence createLeft(Object head, Element[] e, boolean[] drop, boolean foster) {
        return foster
            ? new UnwrapLeft(e, head, drop)
            : new RewritingSequence(head, e, drop);
    }

    /** return a new sequence identical to this one, but with a positive conjunct <tt>s</tt> */
    public Sequence and(Sequence s) {
        if (s.in_a_union)
            throw new RuntimeException("you may not use a sequence as a conjunct if it belongs to a Union");
        Sequence ret = dup();
        ret.needs.add(s);
        s.needed_or_hated=true;
        return ret;
    }

    /** return a new sequence identical to this one, but with a negative conjunct <tt>s</tt> */
    public Sequence andnot(Sequence s) {
        if (s.in_a_union)
            throw new RuntimeException("you may not use a sequence as a conjunct if it belongs to a Union");
        Sequence ret = dup();
        ret.hates.add(s);
        s.needed_or_hated=true;
        return ret;
    }

    /** return a new sequence identical to this one, but with a follow-set restricted to <tt>a</tt> */
    public Sequence followedBy(Atom a) { Sequence ret = dup(); ret.follow = a; return ret; }

    ////////////////////////////////////////////////////////////////////////////////

    abstract Sequence _clone();
    private Sequence dup() {
        Sequence ret = _clone();
        for(Sequence s : needs) { ret.needs.add(s); }
        for(Sequence s : hates) { ret.hates.add(s); }
        ret.follow = follow;
        return ret;
    }

    Iterable<Sequence> needs() { return needs; }
    Iterable<Sequence> hates() { return hates; }

    Position firstp() { return firstp; }
    Position lastp() { return firstp().last(); }

    public Iterator<Element> iterator()    { return new ArrayIterator<Element>(elements); }
    protected Sequence(Element[] elements) {
        this.elements = elements;
        for(int i=0; i<elements.length; i++)
            if (elements[i]==null)
                throw new RuntimeException("cannot have nulls in a sequence: " + this);
        this.firstp = new Position(0, null);
    }

    abstract Forest epsilonForm(Input.Region loc, Grammar cache);

    protected abstract <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p);


    // Position //////////////////////////////////////////////////////////////////////////////

    /** the imaginary position before or after an element of a sequence; corresponds to an "LR item" */
    class Position implements IntegerMappable {

        public int ord = -1;

        private Forest zero = null;
        /*
        public Forest zero(Input.Region reg) {
            if (zero != null) return zero;
            if (pos > 0) throw new RuntimeException("Position.zero(): pos>0");
            return zero = rewrite(reg);
        }
        */
                final int      pos;
        private final Position next;
        private final Position prev;
                final Forest[] holder;
        
        private Position(int pos, Position prev) {
            this.pos      = pos;
            this.next     = pos==elements.length ? null : new Position(pos+1, this);
            this.holder   = new Forest[elements.length];
            this.prev     = prev;
        }

        boolean isFirst() { return pos==0; }

        /** the element immediately after this Position, or null if this is the last Position */
        public Element  element() { return pos>=elements.length ? null : elements[pos]; }

        /** the element which produces the sequence to which this Position belongs */
        public Sequence owner() { return Sequence.this; }

        /** the next Position (the Position after <tt>this.element()</tt>) */
        public Position next() { return next; }

        /** true iff this Position is the last one in the sequence */
        public boolean isLast() { return next()==null; }
        public Position last() { return isLast() ? this : next().last(); }
        public Position prev() { return prev; }

        // Position /////////////////////////////////////////////////////////////////////////////////

        final <T> Forest<T> rewrite(Input.Region loc, Grammar cache) {
            if (this==firstp()) epsilonForm(loc, cache);
            for(int i=0; i<pos; i++) if (holder[i]==null) throw new Error("realbad " + i);
            for(int i=pos; i<elements.length; i++) {
                if (holder[i]==null) holder[i] = ((Union)elements[i]).epsilonForm(loc, cache);
                if (holder[i]==null) throw new Error("bad");
            }
            return Sequence.this.postReduce(loc, holder, this);
        }

        public String   toString() {
            StringBuffer ret = new StringBuffer();
            ret.append("<{");
            for(Position p = Sequence.this.firstp(); p != null; p = p.next()) {
                ret.append(' ');
                if (p==this) ret.append(" | ");
                if (p.element()!=null) ret.append(p.element());
                else                   ret.append(' ');
            }
            ret.append("}>");
            return ret.toString();
        }
        private final int idx = master_position_idx++;
        public int toInt() { return idx; }
    }
    private static int master_position_idx = 0;

    // toString //////////////////////////////////////////////////////////////////////////////

    public String toString() { return toString(new StringBuffer(), false).toString(); }
    StringBuffer toString(StringBuffer sb) { return toString(sb, true); }
    StringBuffer toString(StringBuffer sb, boolean spacing) {
        for(int i=0; i<elements.length; i++) {
            sb.append(elements[i]+"");
            sb.append(' ');
        }
        if (follow != null) {
            sb.append("-> ");
            sb.append(follow);
        }
        for(Sequence s : needs) {
            sb.append("& ");
            sb.append(s);
        }
        for(Sequence s : hates) {
            sb.append("&~ ");
            sb.append(s);
        }
        return sb;
    }


    // Specialized Subclasses //////////////////////////////////////////////////////////////////////////////

    static class Constant extends Sequence {
        private final Object result;
        public Constant(Element[] e, Object result) {
            super(e);
            if (result==null) throw new Error("constant sequences may not have result==null");
            this.result = result;
        }
        Sequence _clone() { return new Constant(elements, result); }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            return (Forest<T>)Forest.create(loc, result, null, false);
        }
        Forest epsilonForm(Input.Region loc, Grammar cache) {
            return Forest.create(loc, result, null, false);
        }
    }

    static class Singleton extends Sequence {
        private final int idx;
        public Singleton(Element e) { this(new Element[] { e }, 0); }
        public Singleton(Element[] e, int idx) { super(e); this.idx = idx; }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) { return args[idx]; }
        Sequence _clone() { return new Singleton(elements,idx); }
        Forest epsilonForm(Input.Region loc, Grammar cache) {
            return ((Union)elements[idx]).epsilonForm(loc, cache);
        }
    }

    static class Unwrap extends Sequence {
        private boolean[] drops;
        private final Object tag;
        public Unwrap(Element[] e, Object tag)                  { this(e, tag, null); }
        public Unwrap(Element[] e, Object tag, boolean[] drops) { super(e); this.drops = drops; this.tag = tag; }
        Sequence _clone() { return new Unwrap(elements, tag, drops); }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            for(int i=0; i<args.length; i++) if (args[i]==null) throw new Error();
            if (drops==null) return Forest.create(loc, (T)tag, args, true);
            int count = 0;
            for(int i=0; i<drops.length; i++) if (!drops[i]) count++;
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, (T)tag, args2, true);
        }
        Forest epsilonForm(Input.Region loc, Grammar cache) {
            return Forest.create(loc, tag, new Forest[0], false);
        }
    }

    static class UnwrapLeft extends Sequence {
        private boolean[] drops;
        private final Object tag;
        public UnwrapLeft(Element[] e, Object tag)                  { this(e, tag, null); }
        public UnwrapLeft(Element[] e, Object tag, boolean[] drops) { super(e); this.drops = drops; this.tag = tag; }
        Sequence _clone() { return new UnwrapLeft(elements, tag, drops); }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            for(int i=0; i<args.length; i++) if (args[i]==null) throw new Error();
            if (drops==null) return Forest.create(loc, (T)tag, args, false, true);
            int count = 0;
            for(int i=0; i<drops.length; i++) if (!drops[i]) count++;
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, (T)tag, args2, false, true);
        }
        Forest epsilonForm(Input.Region loc, Grammar cache) {
            return Forest.create(loc, tag, new Forest[0], false);
        }
    }

    static class RewritingSequence extends Sequence {
        private Object tag;
        private final boolean[] drops;
        private int count = 0;
        Sequence _clone() { return new RewritingSequence(tag, elements, drops); }
        public RewritingSequence(Object tag, Element[] e) { this(tag, e, null); }
        public RewritingSequence(Object tag, Element[] e, boolean[] drops) {
            super(e);
            if (tag==null) throw new Error();
            this.tag = tag;
            this.drops = drops == null ? new boolean[e.length] : drops;
            for(int i=0; i<this.drops.length; i++) if (!this.drops[i]) count++;
        }
        public <T> Forest<T> postReduce(Input.Region loc, Forest<T>[] args, Position p) {
            Forest<T>[] args2 = new Forest[count];
            int j = 0;
            for(int i=0; i<args.length; i++) if (!drops[i]) args2[j++] = args[i];
            return Forest.create(loc, (T)tag, args2, false);
        }
        public StringBuffer toString(StringBuffer sb, boolean spacing) {
            int len = sb.length();
            if (tag != null)
                sb.append("\""+StringUtil.escapify(tag.toString(),"\"\r\n")+"\":: ");
            super.toString(sb, spacing);
            len = sb.length()-len;
            if (spacing) for(int i=0; i<50-len; i++) sb.append(' ');
            return sb;
        }
        Forest epsilonForm(Input.Region loc, Grammar cache) {
            return Forest.create(loc, tag, new Forest[0], false);
        }
    }

}
